#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>
#include <assert.h>

#define MASS(a, i) ((a)[(i) * 5])
#define X(a, i) ((a)[(i) * 5 + 1])
#define Y(a, i) ((a)[(i) * 5 + 2])
#define VX(a, i) ((a)[(i) * 5 + 3])
#define VY(a, i) ((a)[(i) * 5 + 4])

double after, before;

double MPI_Wtime();
double MPI_Wtick();

void supercon2k1_get_config(int problem_num, int * size_p, 
			    int * num_steps_p, double * dt_p);
void supercon2k1_get_data(int problem_num, double * a);
void supercon2k1_verify(int problem_num, double * a);

static struct timeval start_time, end_time;
/*  static char * config_file_name = "problem.dat"; */
/*  static void read_config_file(); */
/*  static int get_param_value(char *param_name, double *value_p); */
static void gen_uniform_square(double *a);
static void gen_galaxy(double *a);
static void gen_uniform_circle(double *a);
static void gen_circular_movement(double *a);
/*static void gen_double_circular_movement(double *a);*/
static void gen_double_circular_movement_2(double *a);
static void gen_circle(double *a);
static void gen_smile(double *a);
static void check_answer(double *correct_data_x, double *correct_data_y,
			 double *correct_data_vx, double *correct_data_vy,
			 double *checked_data, double eps);
void init_rand();
double drand();


static double max_m, min_m;
static double max_v, min_v;
/*  static double ratio = 100.0; */
static int num_particles;
static double dt;

extern double correct_data_0_x [];
extern double correct_data_0_y [];
extern double correct_data_0_vx [];
extern double correct_data_0_vy [];

extern double correct_data_1_x [];
extern double correct_data_1_y [];
extern double correct_data_1_vx [];
extern double correct_data_1_vy [];

void supercon2k1_get_config(int problem_num, int * size_p, 
			    int * num_steps_p, double * dt_p)
{
  switch(problem_num) {
  case 0:
    *size_p = 1024;
    num_particles = *size_p;
    *num_steps_p = 100;
    *dt_p = 0.001;
    break;
  case 1:
    *size_p = 5000;
    num_particles = *size_p;
    *num_steps_p = 100;
    *dt_p = 0.001;
    break;
  case 2:
    *size_p = 20000;
    num_particles = *size_p;
    *num_steps_p = 200;
    *dt_p = 0.05;
    break;
  default:
    fprintf(stderr, "error: unknown problem number: %d\n", problem_num);
    exit(1);
  }
}


void supercon2k1_get_data(int problem_num, double * a)
{
  init_rand(problem_num);
  switch(problem_num) {
  case 0:
    gen_uniform_square(a);
    break;
  case 1:
    gen_uniform_circle(a);
    break;
  case 2:
    gen_circle(a);
    break;
  default:
    fprintf(stderr, "error: unknown problem number: %d\n", problem_num);
    exit(1);
  } 

  gettimeofday(&start_time, NULL);
  
}

static void gen_uniform_circle(double *a)
{
  int i;
  double max_radius;
  double radian;
  double radius;
  double v;

  min_m = 50;
  max_m = 100;
  min_v = 5;
  max_v = 15;
  max_radius = 7.0e2;
  for(i = 0; i < num_particles; i++) {
    MASS(a, i) = min_m + (max_m - min_m) * drand();
    radian = 2 * M_PI * drand();
    radius = max_radius * drand() * drand();
    X(a, i) = radius * cos(radian);
    Y(a, i) = radius * sin(radian);
    radian = 2 * M_PI * drand();
    v = min_v + (max_v - min_v) * drand();
    VX(a, i) = v * cos(radian);
    VY(a, i) = v * sin(radian);
  }
}

static void gen_uniform_square(double *a)
{
  int i, j;
  double constant_energy;
  double v;
  double radian;
  double mean_v;

  min_m = 500;
  max_m = 1000;
  min_v = 100;
  max_v = 200;

  for(i = 0; i < num_particles; i++) {
    double m_i = min_m + (max_m - min_m) * drand();
    double width = 1.0e3;
    MASS(a, i) = m_i;
    X(a, i) = -width  + width  * 2 * drand(); 
    Y(a, i) = -width  + width  * 2 * drand();
  }
  
  /* set same energies to all particles */
  mean_v = min_v + (max_v - min_v) * drand();
  constant_energy = MASS(a, 0) * mean_v * mean_v / 2.0;
  for(i = 1; i < num_particles; i++) {
    double len = (X(a, i) - X(a, 0)) * (X(a, i) - X(a, 0)) 
        + (Y(a, i) - Y(a, 0)) * (Y(a, i) - Y(a, 0)); 
    len = sqrt(len);
    assert(len);
    constant_energy -= MASS(a, 0) * MASS(a, i) / len;     
  }
  radian = M_PI * 2 * drand();
  VX(a, 0) = mean_v * cos(radian);
  VY(a, 0) = mean_v * sin(radian);

  for(i = 1; i < num_particles; i++) {
    double p = constant_energy;
    for(j = 0; j < num_particles; j++) {
      double len;
      if(i == j) continue;
      len = (X(a, i) - X(a, j)) * (X(a, i) - X(a, j))
	+ (Y(a, i) - Y(a, j)) * (Y(a, i) - Y(a, j));
      len = sqrt(len);
      assert(len);
      p += MASS(a, j) * MASS(a, i) / len;     
    }
    v = sqrt(p * 2 / MASS(a, i));
    radian = M_PI * 2 * drand();
    VX(a, i) = v * cos(radian);
    VY(a, i) = v * sin(radian);
  }
}

static void gen_circle(double *a)
{
  int i;
  double radius;
  double radian;
  double max_radius;
  double v;
  
  min_v = 1000;
  max_v = 1500;
  min_m = 1500;
  max_m = 2500;
  max_radius = 1.0e4;
  for(i = 0; i < num_particles; i++) {
    MASS(a, i) = min_m + (max_m - min_m) * drand();
    radius = max_radius * drand() * drand();
    radian = M_PI * 2 * drand();
    X(a, i) = radius * cos(radian);
    Y(a, i) = radius * sin(radian);
    v = min_v + (max_v - min_v) * drand();
    VX(a, i) = v * cos(radian + M_PI_2) * pow(-1, i);
    VY(a, i) = v * sin(radian + M_PI_2) * pow(-1, i);
    v = v * (max_radius - radius) / max_radius + 10;
    VX(a, i) += v * cos(radian);
    VY(a, i) += v * sin(radian);
  }
}

void supercon2k1_verify(int problem_num, double * a)
{
  double elapsed_time_sec;
  gettimeofday(&end_time, NULL);
  elapsed_time_sec = (double)(end_time.tv_sec - start_time.tv_sec)
    + ((double)(end_time.tv_usec - start_time.tv_usec)) / 1.0e6;
  printf("elapsed time [s]: %e\n", elapsed_time_sec);

  switch(problem_num){
  case 0:
    check_answer(correct_data_0_x, correct_data_0_y,
		 correct_data_0_vx, correct_data_0_vy,
		 a, 1.0e-10);
    break;
  case 1: 
    check_answer(correct_data_1_x, correct_data_1_y, 
  		 correct_data_1_vx, correct_data_1_vy, 
  		 a, 1.0e-10);
    break;
/*    case 2: */
/*      check_answer(correct_data_2_x, correct_data_2_y, */
/*  		 correct_data_2_vx, correct_data_2_vy, */
/*  		 a, 1.0e-10); */
/*      break; */
  default: 
    fprintf(stderr, "error: unknown problem number: %d\n", problem_num); 
    exit(1); 
  } 
}

static void check_answer(double *correct_data_x, double *correct_data_y,
			 double *correct_data_vx, double *correct_data_vy,
			 double *checked_data, double eps)
{
  int i = 0;
  int num_verified_data = 0;
  double x_diff, y_diff;
  double vx_diff, vy_diff;

  printf("checking answer...\n");
  for(i = 0; i < num_particles; i++){
    x_diff = fabs(X(checked_data, i) - correct_data_x[i]);
    y_diff = fabs(Y(checked_data, i) - correct_data_y[i]);
    vx_diff = fabs(VX(checked_data, i) - correct_data_vx[i]);
    vy_diff = fabs(VY(checked_data, i) - correct_data_vy[i]);
    if(x_diff <= fabs(eps * correct_data_x[i]) &&
       y_diff <= fabs(eps * correct_data_y[i]) &&
       vx_diff <= fabs(eps * correct_data_vx[i]) &&
       vy_diff <= fabs(eps * correct_data_vy[i])){
      num_verified_data++;
    } else {
      if(x_diff > fabs(eps * correct_data_x[i])){
	printf("x[%d] not verified: %.3e\n", i, 
	       x_diff - eps * correct_data_x[i]);
      }
      if(y_diff > fabs(eps * correct_data_y[i])){
	printf("y[%d] not verified: %.3e\n", i,
	       y_diff - eps * correct_data_y[i]);
      }
      if(vx_diff > fabs(eps * correct_data_vx[i])){
	printf("vx[%d] not verified: %.3e\n", i, 
	       vx_diff - eps * correct_data_vx[i]);
      }
      if(vy_diff > fabs(eps * correct_data_vy[i])){
	printf("vy[%d] not verified: %.3e\n", i,
	       vy_diff - eps * correct_data_vy[i]);
      }
    }
  }
  printf("%d/%d verified\n", num_verified_data, num_particles);
}
